--- flua-demos.lua - some sample flua programs (or almost).

-- Copyright (C) 2001 Eduardo Ochs.
-- Author:     Eduardo Ochs <edrx@mat.puc-rio.br>
-- Version:    0.02 (2001apr29).
--
-- This program was copylefted to prevent against patent psychopaths;
-- if you want a version with any other license you'll have to write
-- it yourself. More formally,
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.

-- Flua is an experimental Forth-like language, built on top of Lua, C
-- and Nasm. It is an implementation of the ideas of my (old) "Crim"
-- project.
--
-- Related files:
-- (find-flua "flua.lua")
-- (find-flua "flua-demos.lua")
-- (find-flua "flua-lua.lua")
-- (find-flua "inc.lua")




-- .flua_demo1_old	(to "flua_demo1_old")
-- .flua_demo1	(to "flua_demo1")
-- .flua_demo2	(to "flua_demo2")
-- .flua_demo3	(to "flua_demo3")
-- .flua_demo4	(to "flua_demo4")
-- .flua_demo5	(to "flua_demo5")

-- (find-fluafile "")
-- (find-angg "CRIM1/crim1a.tcl" "demos")

action["%"]  = function()  tick(getword()) end
action["g%"] = function() gtick(getword()) end
action["\\"] = function() local _; _, _, rest = strfind(rest, "\n(.*)"); end
action["--"] = action["\\"]



-- flua_demo1_old  (to ".flua_demo1_old")
-- This function is here for historical reasons only. It is not invoked.
--
-- `flua_demo1_old' generates the same bytecode and engine file that
-- `flua_demo1' (below) does, but it uses the "old style" calls to do
-- that... From version 0.01 to version 0.02 the function `f', that
-- "compiles" a string to a Flua program, became much more powerful,
-- and each demo in modern form is a single call to `f', with an
-- entire program as its parameter. `f' now even allows recursion, so
-- some words can be macros, or complex Lua functions that call `f'
-- again.
--
-- Most of the demos below have links to their "old style" versions in
-- flua-0.01, and to the bytecode file that flua.lua generates from
-- them (or to the corresponding .lst; the bytecode files are just
-- Nasm code), to the engine file in C and to the output of running
-- them dumping the stacks at every step (the demo<n>.dbg.out files).
--
-- The links will work if you are using a perfectly-configured  ;)
-- Emacs with the eev package installed (see http://angg.twu.net/), or
-- if you are reading the htmlized version of this file, at:
--
--   http://angg.twu.net/LUA/lua-0.02/flua-demos.lua.html
--
-- Questions and comments are welcome. I'm edrx@inx.com.br. 
--
-- (find-angg "LUA/flua-0.01/flua-demos.lua" "flua_demo1")
function flua_demo1_old()
  tick("SQUARE"); f(": DUP * ;")			-- SQUARE ( n -- n^2 )
  tick("CUBE");   f(": DUP SQUARE * ;")			-- CUBE   ( n -- n^3 )
  tick("2");      f("CON");  nasm(4, "\tdd 2\n")	-- 2      ( -- 2 )
  gtick("DEMO");  f(": 2 CUBE ;")			-- DEMO   ( -- )
end




-- flua_demo1  (to ".flua_demo1")
-- (find-flua "flua-demos-old.lua" "flua_demo1_old")
-- (find-fluafile "demo1.dbg.out")
-- (find-fluafile "demo1.bytecode.lst")
-- (find-fluafile "demo1.bytecode.asm")
-- (find-fluafile "demo1.engine.c")
--
function flua_demo1() f[[
  % SQUARE  : DUP * ;			-- SQUARE ( n -- n^2 )
  % CUBE    : DUP SQUARE * ;		-- CUBE   ( n -- n^3 )
  % 2    CON: nasm( 4, "\tdd 2\n");	-- 2      ( -- 2 )
 g% DEMO    : 2 CUBE ; 			-- DEMO   ( -- )
]] end



-- flua_demo2  (to ".flua_demo2")
-- (find-flua "flua-demos-old.lua" "flua_demo2_old")
-- (find-fluafile "demo2.dbg.out")
-- (find-fluafile "demo2.bytecode.lst")
-- (find-fluafile "demo2.bytecode.asm")
-- (find-fluafile "demo2.engine.c")
--
function flua_demo2() f[[
  lua add_SFprims("STO S> TOS >S") end;
  % &FOO   AT:					-- &FOO ( -- adr )
  % FOO!   TO:					-- FOO! ( w -- )
  % FOO   CON: nasm( 4, "\tdd 0x12345\n");	-- FOO  ( -- w )
  % S$@,     : S> COUNT 2DUP + >S ;		-- S$@, ( --
  % <.">  RSR:
  % S<.">    : S$@, TYPE ;
  % 2<."> RSR:
  % S2<.">   : S<."> S<."> ;
 g% DEMO     : 2<."> dbstuff( 6, "Hello\n", 6, "There\n");  FOO FOO + FOO! ;
]] end



-- flua_demo3  (to ".flua_demo3")
-- (find-flua "flua-demos-old.lua" "flua_demo3_old")
-- (find-fluafile "demo3.dbg.out")
-- (find-fluafile "demo3.bytecode.lst")
-- (find-fluafile "demo3.bytecode.asm")
-- (find-fluafile "demo3.engine.c")
--
function flua_demo3() f[[
  % strlen C1: nasm( 4, "extern strlen\n\tdd strlen\n");
  % 0<."> RSR:
  % S0<.">   : S> DUP strlen 2DUP TYPE + 1 + >S ;
 g% DEMO     : 0<."> dbstuff( "Hello\n", 0);
	       0<."> dbstuff( "There!\n", 0); ;
]] end



-- flua_demo4  (to ".flua_demo4")
-- (find-flua "flua-demos-old.lua" "flua_demo4_old")
-- (find-fluafile "demo4.dbg.out")
-- (find-fluafile "demo4.bytecode.lst")
-- (find-fluafile "demo4.bytecode.asm")
-- (find-fluafile "demo4.engine.c")
--
-- (find-flua "flua-comp.lua" "words:if_then_else")
--
function flua_demo4() f[[
  lua c1word("strlen") end;
  % 0<."> RSR:
  % S0<.">   : S> DUP strlen 2DUP TYPE + 1 + >S ;
  function zdotquo (...) f("0<.\">"); call(dbstuff, arg) end;

  % LITW  RSR: : SGOBBLE2 ;
  function litw (n) f("LITW"); nasm(2, format("\tdw %s\n", n)) end;

  % BRANCH  RSR: :  SBRANCH ;
  % 0BRANCH RSR: : S0BRANCH ;
  % YES/NO       : if   zdotquo( "yes\n", 0);
	           else zdotquo( "no\n", 0);
		   then ;
 g% DEMO         : zdotquo( "Hello\n", 0);
	           litw( 22); YES/NO  litw( 0); YES/NO ;
]] end



-- flua_demo5  (to ".flua_demo5")
-- (find-flua "flua-demos-old.lua" "flua_demo5_old")
-- (find-fluafile "demo5.dbg.out")
-- (find-fluafile "demo5.bytecode.lst")
-- (find-fluafile "demo5.bytecode.asm")
-- (find-fluafile "demo5.engine.c")
--
-- (find-flua "flua-lua.lua" "lua_api_words")
-- (find-node "(lua)States" "lua_State *lua_open")
-- (find-node "(lua)Pushing Values Onto The Stack" "lua_pushstring")
-- (find-node "(lua)Calling Lua Functions")
-- lua_getglobal	-- lua_getglobal  ( $adr -- )
-- lua_pushstring	-- lua_pushstring ( $adr -- )
-- lua_call		-- lua_call       ( nargs nresults -- )
--
function flua_demo5()
  f[[

  lua add_luastuff()		-- for several lua_xxx words
      c1word("strlen")          -- strlen ( adr -- len )
  end;
  % 0$ RSR:			-- S0$ ( S: adr -- S: adr+len+1 D: adr )
  % S0$   :  S> DUP DUP strlen + 1 + >S ;
  function zstr (str) f("0$") dbstuff(str, 0) end;

  g% DEMO :  lua_open lua_state! lua_stdlibsopen
             zstr( "print"); lua_getglobal
             zstr( "Hello"); lua_pushstring
             1 0 lua_call
          ;
  ]]
end







-- (find-es "crim" "crim1b")
-- (find-angg "LUA/inc.lua")
-- (find-angg "LUA/engine0.skel.c")
-- (find-angg "LUA/nasmskel.asm")
-- (find-fline "~/tmp/lua-manual.txt")

-- (find-fline "/usr/include/")
-- (find-fline "/usr/include/lua.h")
-- (find-fline "/usr/include/luadebug.h")
-- (find-fline "/usr/include/lualib.h")






--
-- Local Variables:
-- coding:               no-conversion
-- ee-anchor-format:     "%s"
-- ee-charset-indicator: ""
-- End:
