--- flua-lua.lua - Functions to let the Flua engine link with the
--- Lua libs.
--
-- Copyright (C) 2001 Eduardo Ochs.
-- Author:     Eduardo Ochs <edrx@mat.puc-rio.br>
-- Version:    0.02 (2001apr29).
--
-- This program was copylefted to prevent against patent psychopaths;
-- if you want a version with any other license you'll have to write
-- it yourself. More formally,
--
-- This program is free software; you can redistribute it and/or
-- modify it under the terms of the GNU General Public License as
-- published by the Free Software Foundation; either version 2 of the
-- License, or (at your option) any later version.


-- Flua is an experimental Forth-like language, built on top of Lua, C
-- and Nasm. It is an implementation of the ideas of my (old) "Crim"
-- project.
--
-- Related files:
-- (find-flua "flua.lua")
-- (find-flua "flua-comp.lua")
-- (find-flua "flua-demos.lua")
-- (find-flua "flua-demos.lua" "flua_demo5")
-- (find-flua "inc.lua")



-- .lua_stdlibsopen	(to "lua_stdlibsopen")
-- .lua_state_prims	(to "lua_state_prims")
-- .lua_heads		(to "lua_heads")
-- .lua_0heads	(to "lua_0heads")
-- .lua_api_words	(to "lua_api_words")

-- (find-luafile "src/lua/lua.c" "static void userinit (void) {")
-- (find-fline "/usr/include/lualib.h")
-- (find-fline "/usr/doc/lua/etc/trace.c")
-- (find-luafile "loadlib.patch")


-- Currently everything is in one huge function; you call it and it
-- adds lua words to the engine file and to the bytecode file.
-- Examples of how to call it:
--
-- add_luastuff()
-- add_luastuff("lua_rexlibopen(L); lua_bitlibopen(L); lua_loadlib_open(L);")
--
function add_luastuff(extralibs)
  -- lua_stdlibsopen  (to ".lua_stdlibsopen")
  strings.Cextras1 = [[
    #include "lua.h"
    #include "lualib.h"
    void lua_stdlibsopen (lua_State *L) {
      lua_baselibopen(L);
      lua_iolibopen(L);
      lua_strlibopen(L);
      lua_mathlibopen(L);
      lua_dblibopen(L);
      /* extra libraries: */
]] .. (extralibs or "") .. [[
    }
    /* define L (from now on!) as a shorthand for a long variable name: */
    #define L flua_lua_state
    lua_State *L = NULL;
]]

  -- lua_state_prims  (to ".lua_state_prims")
  -- lua_state  ( -- luaS* )
  -- lua_state! ( luaS* -- )
  --
  add_Fprims(
    "lua_state "," DS[1]=(void *)L; DS++; goto forth;",
    "lua_state!"," L=(void *)(DS[0]); DS--; goto forth;"
  )

  -- lua_heads  (to ".lua_heads")
  -- Heads to invoke functions from the Lua libraries.
  -- LUAC0 ( R: ip0 ip  D:          -- R: ip0 D: retval )
  -- LUAC1 ( R: ip0 ip  D: p1       -- R: ip0 D: retval )
  -- LUAC2 ( R: ip0 ip  D: p1 p2    -- R: ip0 D: retval )
  -- LUAC3 ( R: ip0 ip  D: p1 p2 p3 -- R: ip0 D: retval )
  --
  add_Hprims(
  "LUAC0 "," fun=*(funptr *)(_f0+RS[0]); DS[1]=(*fun)(L); RS--;"
         .." DS++; goto forth;",
  "LUAC1 "," fun=*(funptr *)(_f0+RS[0]); DS[0]=(*fun)(L,DS[0]); RS--;"
         .." goto forth;",
  "LUAC2 "," fun=*(funptr *)(_f0+RS[0]); DS[-1]=(*fun)(L,DS[-1],DS[0]);"
         .." DS--; RS--; goto forth;",
  "LUAC3 "," fun=*(funptr *)(_f0+RS[0]); DS[-2]=(*fun)(L,DS[-2],DS[-1],DS[0]);"
         .." DS-=2; RS--; goto forth;"
  )

  -- lua_0heads  (to ".lua_0heads")
  -- The corresponding "composite heads" that drop the retval:
  action["LUA0C0:"] = function() f("DROPPING: LUAC0:") end
  action["LUA0C1:"] = function() f("DROPPING: LUAC1:") end
  action["LUA0C2:"] = function() f("DROPPING: LUAC2:") end
  action["LUA0C3:"] = function() f("DROPPING: LUAC3:") end


  -- Define some Flua words as being calls to the Lua libraries.
  -- Note that "lua_stdlibsopen" is defined in the Flua engine file,
  -- but behaves like a Lua library function.
  -- DECLARING THESE WORDS HERE IS UGLY; THE RIGHT WAY WOULD BE TO
  -- DEVISE A "LAZY" WAY OF DECLARING THEM THAT WOULD PUT IN THE
  -- BYTECODE ONLY THE WORDS THAT ARE USED LATER.
  -- lua_api_words  (to ".lua_api_words")
  --
  -- (find-node "(lua)States")
  --
  cnword("lua_open",        "C0:")	-- lua_open        ( -- lua_State* )
  cnword("lua_close",       "LUA0C0:")	-- lua_close	   ( -- )
  cnword("lua_stdlibsopen", "LUA0C0:")  -- lua_stdlibsopen ( -- )

  -- (find-node "(lua)The Stack and Indices")
  --
  cnword("lua_gettop", "LUAC0:")	-- lua_gettop     ( -- nelements )
  cnword("lua_stackspace", "LUAC0:")	-- lua_stackspace ( -- nelements )

  -- (find-node "(lua)Stack Manipulation")
  --
  cnword("lua_settop",    "LUA0C1:")		-- lua_settop    ( index -- )
  cnword("lua_pushvalue", "LUA0C1:")		-- lua_pushvalue ( index -- )
  cnword("lua_remove",    "LUA0C1:")		-- lua_remove    ( index -- )
  cnword("lua_insert",    "LUA0C1:")		-- lua_insert    ( index -- )
  add_Fprims("lua_pop",				-- lua_pop   ( nelements -- )
    "lua_settop(L, -(DS[0])-1); DS--; goto forth;")

  -- (find-node "(lua)Pushing Values Onto The Stack")
  --
  add_Fprims("lua_pushnumber",			-- lua_pushnumber ( n -- )
    "lua_pushnumber(L, DS[0]); DS--; goto forth;")
  cnword("lua_pushlstring",   "LUA0C2:")    -- lua_pushlstring   ( adr len -- )
  cnword("lua_pushstring",    "LUA0C1:")    -- lua_pushstring    ( adr -- )
  cnword("lua_pushusertag",   "LUA0C2:")    -- lua_pushusertag   ( adr tag -- )
  cnword("lua_pushnil",       "LUA0C0:")    -- lua_pushnil       ( -- )
  -- cnword("lua_pushcfunction", "LUA0C1:") -- lua_pushcfunction ( luacfun -- )

  -- (find-node "(lua)Executing Lua Code")
  --
  cnword("lua_dofile",   "LUAC1:")	-- lua_dofile   ( fnameadr -- )
  cnword("lua_dostring", "LUAC1:")	-- lua_dostring ( stradr -- )
  cnword("lua_dobuffer", "LUAC3:")	-- lua_dobuffer ( adr len nameadr -- )

  -- (find-node "(lua)Manipulating Global Variables in Lua")
  --
  cnword("lua_getglobal",  "LUA0C1:")		-- lua_getglobal ( varname -- )
  cnword("lua_setglobal",  "LUA0C1:")		-- lua_setglobal ( varname -- )
  cnword("lua_getglobals", "LUA0C0:")		-- lua_getglobals ( -- )
  cnword("lua_setglobals", "LUA0C0:")		-- lua_setglobals ( -- )

  -- (find-node "(lua)Manipulating Tables in Lua")
  --
  cnword("lua_gettable", "LUA0C1:")		-- lua_gettable ( index -- )
  cnword("lua_rawget",   "LUA0C1:")		-- lua_rawget   ( index -- )
  cnword("lua_settable", "LUA0C1:")		-- lua_settable ( index -- )
  cnword("lua_rawset",   "LUA0C1:")		-- lua_rawset   ( index -- )
  cnword("lua_newtable", "LUA0C0:")		-- lua_newtable ( -- )

  -- (find-node "(lua)Using Tables as Arrays")
  --
  cnword("lua_rawgeti", "LUA0C2:")		-- lua_rawgeti ( index n -- )
  cnword("lua_rawseti", "LUA0C2:")		-- lua_rawseti ( index n -- )
  cnword("lua_getn",    "LUAC1:")		-- lua_getn    ( index -- n)

  -- (find-node "(lua)Calling Lua Functions")
  --
  cnword("lua_call",    "LUAC2:")     -- lua_call ( nargs nresults -- errcode )
  cnword("lua_rawcall", "LUA0C2:")	 -- lua_rawcall ( nargs nresults -- )
  cnword("lua_error",   "LUA0C1:")	 -- lua_error        ( msgstr -- )
  cnword("lua_settagmethod", "LUA0C2:")  -- lua_settagmethod ( tag evname -- )
  cnword("lua_gettagmethod", "LUA0C2:")  -- lua_gettagmethod ( tag evname -- )
  cnword("lua_copytagmethods", "LUAC2:") -- lua_copytagmethods ( tto tfrom-- )
  cnword("lua_next",   "LUAC1:")	 -- lua_next       ( index -- status )
  cnword("lua_concat", "LUA0C1:")	 -- lua_concat     ( n -- )
end





-- (find-luafile "include/lua.h")

--
-- Local Variables:
-- coding:               no-conversion
-- ee-anchor-format:     "%s"
-- ee-charset-indicator: ""
-- End:
